/*****************************************************************************
 * Copyright © 2011 - 2013, M2X BV
 * Author: Jean-Paul Saman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ****************************************************************************/

#ifndef VLC_VAAPI_H
#define VLC_VAAPI_H 1

#ifdef HAVE_AVCODEC_VAAPI

/* VA API connection shared structure
 * The decoding thread and the display thread
 * need to share the connection to libva and
 * the created surfaces for use with decoding
 * context.
 */
typedef struct vlc_va_conn_t vlc_va_conn_t;
typedef struct picture_sys_t vlc_va_surface_t;

struct vlc_va_conn_t
{
    Display      *x11; /* x11 display connection */
    VADisplay     p_display;

    int           i_version_major;
    int           i_version_minor;
    int           i_ref_count;

    VAContextID   i_context_id;
    VAImageFormat derive_format;

    bool          b_direct_rendering; /* VAAPI uses direct rendering */
    bool          b_supports_derive;  /* supports vaDerive */

    /* NOTE: must be called with conn->lock held */
    struct {
        unsigned int count;   /* number of surfaces allocated */
        unsigned int current; /* current used surface (private use only) */
        vlc_va_surface_t *p_surfaces;
    } pool; /* vlc_va_surface_t pool */

    /* locking functions */
    void (*lock)(void);
    void (*unlock)(void);

    /* surface functions */
    bool (*create_surfaces)(vlc_va_conn_t *conn, const int width, const int height,
                                   const int fourcc, const unsigned num);
    void (*destroy_surfaces)(vlc_va_conn_t *conn);

    vlc_va_surface_t *(*get_surface)(vlc_va_conn_t *conn);
    void (*release_surface)(vlc_va_conn_t *conn, vlc_va_surface_t *surface);
    void (*release_surface_by_id)(vlc_va_conn_t *conn, VASurfaceID id);
};

/* Initialize shared connection to libva */
vlc_va_conn_t *vlc_va_Initialize(const char *display);

/* Deinitialize shared connection to libva */
void vlc_va_Terminate(vlc_va_conn_t *conn);

static inline bool vlc_va_conn_CreateSurface(vlc_va_conn_t *conn,
    const int width, const int height, const int fourcc, const int num)
{
    return conn->create_surfaces(conn, width, height, fourcc, num);
}

static inline void vlc_va_conn_DestroySurfaces(vlc_va_conn_t *conn)
{
    conn->destroy_surfaces(conn);
}

static inline vlc_va_surface_t *vlc_va_conn_GetSurface(vlc_va_conn_t *conn)
{
    return conn->get_surface(conn);
}

static inline void vlc_va_conn_ReleaseSurface(vlc_va_conn_t *conn, vlc_va_surface_t *surface)
{
    conn->release_surface(conn, surface);
}

static inline void vlc_va_conn_ReleaseSurfaceById(vlc_va_conn_t *conn, VASurfaceID surface_id)
{
    conn->release_surface_by_id(conn, surface_id);
}

static inline unsigned int vlc_va_conn_SurfacesCount(vlc_va_conn_t *conn)
{
    return conn->pool.count;
}

/* Lock libva connection before use */
int vlc_va_get_video_format( const VAImageFormat *format, video_format_t *fmt );
int vlc_va_get_image_format(vlc_va_conn_t *conn, const unsigned int fourcc,
                            VAImageFormat *format, video_format_t *fmt);

/* picture_sys_t */
struct picture_sys_t
{
    VASurfaceID  i_id;
    int          i_refcount;
    unsigned int i_cache;    /* subpicture cache identifier */

    vlc_mutex_t  lock;
};

static inline void vlc_va_SurfaceHold(vlc_va_surface_t *surface)
{
    vlc_mutex_lock(&surface->lock);
    surface->i_refcount++;
    vlc_mutex_unlock(&surface->lock);
}
static inline void vlc_va_SurfaceRelease(vlc_va_surface_t *surface)
{
    vlc_mutex_lock(&surface->lock);
    surface->i_refcount--;
    vlc_mutex_unlock(&surface->lock);
}

#endif
#endif
