/**
 * @file vaapi/glx.c
 * @brief VA API using GLX video output module for VLC media player
 */
/*****************************************************************************
 * Copyright © 2013, M2X BV
 * Author: Jean-Paul Saman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <assert.h>

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_es.h>
#include <vlc_picture.h>
#include <vlc_vout_display.h>
#include <vlc_vout_window.h>
#include <vlc_picture_pool.h>

#include <vlc_arrays.h>

#ifdef HAVE_LIBAVCODEC_AVCODEC_H
#   include <libavcodec/avcodec.h>
#   include <libavcodec/vaapi.h>
#else
#   include <avcodec.h>
#   include <vaapi.h>
#endif

#include "../../codec/avcodec/avcodec.h"

#include <xcb/xcb.h>
#include <X11/Xlib-xcb.h>
#include <X11/Xutil.h>
#include <vlc_xlib.h>
#include "../../video_output/xcb/xcb_events_vlc.h"

#include <GL/glx.h>
#include <GL/glxext.h>
#include <vlc_opengl.h>

#include <va/va.h>
#include <va/va_x11.h>
#include <va/va_glx.h>
#include "../../codec/avcodec/va.h"
#include "../../codec/avcodec/vaapi.h"

#include "spu.h"
#include "common.h"
#include "xcb.h"
#include "glx.h"

#define VOUT_CFG_PREFIX "vaapi-vout-"

/* define for extra debugging */
#undef VAAPI_DEBUG

static picture_pool_t *Pool(vout_display_t *, unsigned);
static void Render(vout_display_t *, picture_t *, subpicture_t *);
static void DisplayPicture(vout_display_t *, picture_t *, subpicture_t *);

/* */
struct vout_display_sys_t
{
    vlc_va_conn_t    *va;
    xcb_connection_t *conn; /**< XCB connection */

    vout_window_t *embed;

    picture_pool_t *pool;  /* picture pool */
    VAImageFormat img_fmt;
    VAImageFormat sub_fmt; /* Subpicture format VA_FOURCC_RGBA */
    int           sflags;  /* Subtitle format flags */

    struct {
        unsigned int   i_cache;
        mtime_t        i_start;
        mtime_t        i_stop;
        subpicture_t  *prev_subpicture;
    } render;

    vlc_va_spu_cache_t *spu;

    xcb_cursor_t cursor; /* blank cursor */
    xcb_window_t window; /* drawable X window */
    xcb_window_t glwin;  /* GLX window */

    /* OpenGL */
    vlc_gl_t gl;
    GLXContext ctx;
    GLuint gl_texture;
    void *gl_surface;

    const vlc_chroma_description_t *chroma;

    bool visible; /* whether to draw */
};

static bool CheckGLXversion(vout_display_t *vd, Display *dpy, int major, int minor)
{
    int i_major, i_minor;

    if (!glXQueryVersion(dpy, &i_major, &i_minor))
    {
        msg_Dbg (vd, "GLX extension not available");
        return false;
    }

    if (i_major != major)
        msg_Dbg (vd, "GLX extension version %d.%d unknown", i_major, i_minor);
    else if (i_minor < minor)
        msg_Dbg (vd, "GLX extension version %d.%d too old", i_major, i_minor);
    else
    {
        msg_Dbg (vd, "using GLX extension version %d.%d", i_major, i_minor);
        return (i_minor >= minor);
    }

    return false;
}

static void SwapBuffers(vlc_gl_t *gl)
{
    vout_display_sys_t *sys = gl->sys;

    glXSwapBuffers(sys->va->x11, sys->glwin);
}

static void *GetProcAddress(vlc_gl_t *gl, const char *name)
{
    (void)gl;
#ifdef GLX_ARB_get_proc_address
    return glXGetProcAddressARB ((const GLubyte *)name);
#else
    return NULL;
#endif
}

static inline bool HasExtension(const char *apis, const char *api)
{
    size_t apilen = strlen(api);
    while (apis) {
        while (*apis == ' ')
            apis++;
        if (!strncmp(apis, api, apilen) && memchr(" ", apis[apilen], 2))
            return true;
        apis = strchr(apis, ' ');
    }
    return false;
}

static int InitializeGLXv13(vout_display_t *vd)
{
    vout_display_sys_t *sys = vd->sys;
    vlc_va_conn_t *va = sys->va;
    Display *dpy = va->x11;

    /* Find window parameters */
    unsigned snum;
    uint8_t depth;
    uint16_t width, height;
    const xcb_screen_t *scr = FindWindow(vd, sys->embed,
                                         sys->conn, &snum, &depth,
                                         &width, &height);
    if (scr == NULL)
        return VLC_EGENERIC;

    sys->window = xcb_generate_id(sys->conn);

    /* GLX 1.3 */
    static const int attr[] = {
        GLX_RED_SIZE, 5,
        GLX_GREEN_SIZE, 5,
        GLX_BLUE_SIZE, 5,
        GLX_DOUBLEBUFFER, True,
        GLX_X_RENDERABLE, True,
        GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
        None };

    xcb_get_window_attributes_reply_t *wa =
            xcb_get_window_attributes_reply(sys->conn,
                                            xcb_get_window_attributes(sys->conn, sys->embed->handle.xid),
                                            NULL);
    if (wa == NULL)
        return VLC_EGENERIC;
    xcb_visualid_t visual = wa->visual;
    free(wa);

    int nelem;
    GLXFBConfig *confs = glXChooseFBConfig(dpy, snum, attr, &nelem);
    if (confs == NULL)
    {
        msg_Err (vd, "no GLX frame buffer configurations");
        return VLC_EGENERIC;
    }

    GLXFBConfig conf;
    bool found = false;

    for (int i = 0; i < nelem && !found; i++)
    {
        conf = confs[i];

        XVisualInfo *vi = glXGetVisualFromFBConfig(dpy, conf);
        if (vi == NULL)
            continue;

        if (vi->visualid == visual)
            found = true;
        XFree(vi);
    }
    XFree(confs);

    if (!found)
    {
        msg_Err (vd, "no matching GLX frame buffer configuration");
        return VLC_EGENERIC;
    }

    sys->glwin = None;
    if (!CreateWindow(vd, sys->embed, sys->conn, depth, 0 /* ??? */,
                      width, height, &sys->window))
        sys->glwin = glXCreateWindow(dpy, conf, sys->window, NULL );
    if (sys->glwin == None)
    {
        msg_Err (vd, "cannot create GLX window");
        return VLC_EGENERIC;
    }

    sys->cursor = CreateBlankCursor(sys->conn, scr);

    /* Create an OpenGL context */
    sys->ctx = glXCreateNewContext(dpy, conf, GLX_RGBA_TYPE, NULL, True);
    if (sys->ctx == NULL)
    {
        msg_Err (vd, "cannot create GLX context");
        return VLC_EGENERIC;
    }

    if (!glXMakeContextCurrent(dpy, sys->glwin, sys->glwin, sys->ctx))
        return VLC_EGENERIC;

    const char *glx_extensions = glXQueryExtensionsString(dpy, snum);

    bool is_swap_interval_set = false;
#ifdef GLX_SGI_swap_control
    if (HasExtension(glx_extensions, "GLX_SGI_swap_control")) {
        PFNGLXSWAPINTERVALSGIPROC SwapIntervalSGI = (PFNGLXSWAPINTERVALSGIPROC)GetProcAddress(NULL, "glXSwapIntervalSGI");
        if (!is_swap_interval_set && SwapIntervalSGI)
            is_swap_interval_set = !SwapIntervalSGI (1);
    }
#endif
#ifdef GLX_EXT_swap_control
    if (HasExtension(glx_extensions, "GLX_EXT_swap_control")) {
        PFNGLXSWAPINTERVALEXTPROC SwapIntervalEXT = (PFNGLXSWAPINTERVALEXTPROC)GetProcAddress(NULL, "glXSwapIntervalEXT");
        if (!is_swap_interval_set && SwapIntervalEXT)
        {
            SwapIntervalEXT(dpy, sys->glwin, 1);
            is_swap_interval_set = true;
        }
    }
#endif

    return VLC_SUCCESS;
}

static int CreateGLXSurface(vout_display_t *vd)
{
    vout_display_sys_t *sys = vd->sys;
    vlc_va_conn_t *va = sys->va;

    /* Common GLX config */
    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);
    glDisable(GL_CULL_FACE);

    glDrawBuffer(GL_BACK); /* double buffered */

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);

    /* Create OpenGL texture */
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &sys->gl_texture);

    glBindTexture(GL_TEXTURE_2D, sys->gl_texture);
    if (glIsTexture(sys->gl_texture) == GL_FALSE)
    {
        msg_Err(vd, "unable to allocate a GL texture");
        sys->gl_texture = GL_NONE;
        return VLC_EGENERIC;
    }

    /* Set the texture parameters */
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_PRIORITY, 1.0);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                 vd->fmt.i_width, vd->fmt.i_height, 0,
                 GL_BGRA, GL_UNSIGNED_BYTE, NULL);

    glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);

    /* Create GLX surfaces */
    va->lock();
    VAStatus status = vaCreateSurfaceGLX(va->p_display,
                                         GL_TEXTURE_2D,
                                         sys->gl_texture,
                                        &sys->gl_surface);
    va->unlock();
    if (status != VA_STATUS_SUCCESS)
    {
        msg_Err(vd, "unable to allocate GLX Surface");
        return VLC_EGENERIC;
    }

    return VLC_SUCCESS;
}

static int Control(vout_display_t *vd, int query, va_list ap)
{
    vout_display_sys_t *sys = vd->sys;

    switch (query)
    {
    case VOUT_DISPLAY_CHANGE_FULLSCREEN:
    {
        const vout_display_cfg_t *c = va_arg(ap, const vout_display_cfg_t *);
        return vout_window_SetFullScreen(sys->embed, c->is_fullscreen);
    }

    case VOUT_DISPLAY_CHANGE_WINDOW_STATE:
    {
        unsigned state = va_arg(ap, unsigned);
        return vout_window_SetState(sys->embed, state);
    }

    case VOUT_DISPLAY_CHANGE_DISPLAY_SIZE:
    case VOUT_DISPLAY_CHANGE_ZOOM:
    case VOUT_DISPLAY_CHANGE_DISPLAY_FILLED:
    case VOUT_DISPLAY_CHANGE_SOURCE_ASPECT:
    case VOUT_DISPLAY_CHANGE_SOURCE_CROP:
    {
        const vout_display_cfg_t *cfg;
        const video_format_t *source;
        bool is_forced = false;

        if (query == VOUT_DISPLAY_CHANGE_SOURCE_ASPECT
         || query == VOUT_DISPLAY_CHANGE_SOURCE_CROP)
        {
            source = (const video_format_t *)va_arg(ap, const video_format_t *);
            cfg = vd->cfg;
        }
        else
        {
            source = &vd->source;
            cfg = (const vout_display_cfg_t*)va_arg(ap, const vout_display_cfg_t *);
            if (query == VOUT_DISPLAY_CHANGE_DISPLAY_SIZE)
                is_forced = (bool)va_arg(ap, int);
        }

        /* */
        if (query == VOUT_DISPLAY_CHANGE_DISPLAY_SIZE
         && is_forced
         && (cfg->display.width  != vd->cfg->display.width
           ||cfg->display.height != vd->cfg->display.height)
         && vout_window_SetSize(sys->embed,
                                cfg->display.width, cfg->display.height))
            return VLC_EGENERIC;

        vout_display_place_t place;
        vout_display_PlacePicture(&place, source, cfg, false);

        /* Move the picture within the window */
        const uint32_t values[] = { place.x, place.y,
                                    place.width, place.height, };
        xcb_void_cookie_t ck =
            xcb_configure_window_checked(sys->conn, sys->window,
                            XCB_CONFIG_WINDOW_X | XCB_CONFIG_WINDOW_Y
                          | XCB_CONFIG_WINDOW_WIDTH | XCB_CONFIG_WINDOW_HEIGHT,
                              values);
        if (CheckError(vd, sys->conn, "cannot resize XCB window", ck))
            return VLC_EGENERIC;

        glViewport(0, 0, place.width, place.height);
        return VLC_SUCCESS;
    }

    /* Hide the mouse. It will be send when
     * vout_display_t::info.b_hide_mouse is false */
    case VOUT_DISPLAY_HIDE_MOUSE:
    {
        xcb_change_window_attributes(sys->conn, sys->embed->handle.xid,
                                     XCB_CW_CURSOR, &(uint32_t){ sys->cursor });
        xcb_flush(sys->conn);
        return VLC_SUCCESS;
    }

    case VOUT_DISPLAY_GET_OPENGL:
    {
        vlc_gl_t **gl = va_arg (ap, vlc_gl_t **);
        *gl = &sys->gl;
        return VLC_SUCCESS;
    }

    case VOUT_DISPLAY_RESET_PICTURES:
        assert(0);
    default:
        msg_Err (vd, "Unknown request in VAAPI vout display");
        return VLC_EGENERIC;
    }
}

static void Manage(vout_display_t *vd)
{
    int canc = vlc_savecancel();
    vout_display_sys_t *sys = vd->sys;
    ManageEvent(vd, sys->conn, &sys->visible);
    vlc_restorecancel(canc);
}

/**
 * Probe the X server.
 */
int OpenVaapiGLX(vlc_object_t *obj)
{
    vout_display_t *vd = (vout_display_t *) obj;

    const bool b_use_hw = var_GetBool(vd->p_libvlc, "ffmpeg-hw");
    if (!b_use_hw)
    {
        msg_Warn(vd, "requires GPU hardware decoding to be enabled (option --ffmpeg-hw)");
        return VLC_EGENERIC;
    }

    if (!vlc_xlib_init(VLC_OBJECT(vd)))
        return VLC_EGENERIC;

    vout_display_sys_t *sys = calloc(1, sizeof(*sys));
    if (unlikely(sys == NULL))
        return VLC_ENOMEM;

    vd->sys = sys;

    sys->spu = vlc_va_spu_cache_New();
    if (!sys->spu)
    {
        free(sys);
        return VLC_ENOMEM;
    }

    sys->render.prev_subpicture = NULL;
    sys->render.i_cache = 0;

    sys->visible = false;

    sys->embed = MakeWindow(vd);
    if (unlikely(sys->embed == NULL))
    {
        CloseVaapiGLX(obj);
        return VLC_EGENERIC;
    }

    /* Create a VA display */
    vlc_va_conn_t *va = vlc_va_Initialize(sys->embed->display.x11);
    if (unlikely(va == NULL))
    {
        CloseVaapiGLX(obj);
        return VLC_EGENERIC;
    }

    /* Connect to X server */
    xcb_connection_t *conn = xcb_connect(sys->embed->display.x11, NULL);
    if (unlikely(xcb_connection_has_error (conn)))
    {
        CloseVaapiGLX(obj);
        return VLC_EGENERIC;
    }

    sys->va = va;
    sys->conn = conn;

    va->lock();

    XSetEventQueueOwner(va->x11, XCBOwnsEventQueue);
    RegisterMouseEvents(obj, conn, sys->embed->handle.xid);

    sys->gl_texture = GL_NONE;
    sys->gl_surface = NULL;
    sys->gl.sys = NULL;
    sys->ctx = NULL;

    if (!CheckGLXversion(vd, va->x11, 1, 3))
    {
        va->unlock();
        CloseVaapiGLX(obj);
        return VLC_EGENERIC;
    }

    if (InitializeGLXv13(vd) != VLC_SUCCESS)
    {
        va->unlock();
        CloseVaapiGLX(obj);
        return VLC_EGENERIC;
    }

    /* Initialize common OpenGL video display */
    sys->gl.lock = NULL;
    sys->gl.unlock = NULL;
    sys->gl.swap = SwapBuffers;
    sys->gl.getProcAddress = GetProcAddress;
    sys->gl.sys = sys;

    video_format_t fmt;
    int result = vlc_va_get_image_format(va, VA_FOURCC( 'Y', 'V', '1', '2' ),
                                         &sys->img_fmt, &fmt);
    va->unlock();

    if (result != VLC_SUCCESS)
    {
        msg_Err(vd, "GPU does not support requested fourcc");
        CloseVaapiGLX(obj);
        return VLC_EGENERIC;
    }

    vd->fmt.i_chroma = fmt.i_chroma;
    vd->fmt.i_bits_per_pixel = fmt.i_bits_per_pixel;

    sys->chroma = vlc_fourcc_GetChromaDescription(vd->fmt.i_chroma);

    /* Initialize common OpenGL video display */
    if (CreateGLXSurface(vd) != VLC_SUCCESS)
    {
        CloseVaapiGLX(obj);
        return VLC_EGENERIC;
    }
#if 1 /* DEBUG CODE */
    msg_Dbg(vd, "video output format size %d x %d (%d x %d)",
            vd->fmt.i_width, vd->fmt.i_height,
            vd->fmt.i_visible_width, vd->fmt.i_visible_height);
#endif
    /* */
    vout_display_info_t info = vd->info;
    info.is_slow = false;
    info.has_pictures_invalid = false;
    info.has_event_thread = true;
    info.has_hide_mouse = false;
    /* NOTE: Needs to be false for AMD/ATI Graphics cards */
    info.can_recycle = var_CreateGetBool(vd, VOUT_CFG_PREFIX "recycle");

    /* Setup vout_display_t once everything is fine */
    if (VASubtitleFourCC(va, VA_FOURCC_RGBA, &sys->sub_fmt, &sys->sflags) == VLC_SUCCESS)
        info.subpicture_chromas = va_subpicture_chromas;
    else
        msg_Warn(vd, "libva does not support fourcc RGBA subtitles");

    /* Initialize common video display */
    vd->info = info;
    vd->pool = Pool;
    vd->manage = Manage;
    vd->control = Control;
    vd->display = DisplayPicture;
    vd->prepare = vd->info.subpicture_chromas ? Render : NULL;

    /* */
    if (!(sys->sflags & VA_SUBPICTURE_GLOBAL_ALPHA) &&
        !(sys->sflags & VA_SUBPICTURE_CHROMA_KEYING))
        msg_Warn(vd, "Display server does not support alpha or chroma for subtitles.");

    /* */
    bool is_fullscreen = vd->cfg->is_fullscreen;
    if (is_fullscreen && vout_window_SetFullScreen(sys->embed, true))
        is_fullscreen = false;
    vout_display_SendEventFullscreen(vd, is_fullscreen);
    vout_display_SendEventDisplaySize(vd, vd->fmt.i_width, vd->fmt.i_height, is_fullscreen);

    msg_Info(obj, "using VAAPI GLX video output (libva version %d.%d)",
             va->i_version_major, va->i_version_minor);

    return VLC_SUCCESS;
}

void CloseVaapiGLX(vlc_object_t *obj)
{
    vout_display_t *vd = (vout_display_t *) obj;
    vout_display_sys_t *sys = vd->sys;
    vlc_va_conn_t *va = (vlc_va_conn_t *)sys->va;

    sys->visible = false;

    /* Cleanup GLX resources */
    if (sys->gl_surface)
    {
        VAStatus status = vaDestroySurfaceGLX(va->p_display, sys->gl_surface);
        if (status != VA_STATUS_SUCCESS)
            msg_Err(vd, "GLX surface not destroyed");
    }

    if (sys->ctx != NULL)
    {
        glXMakeContextCurrent(va->x11, None, None, NULL);
        glXDestroyContext(va->x11, sys->ctx);
        if (sys->glwin)
            glXDestroyWindow(va->x11, sys->glwin);
    }

    if (sys->gl_texture != GL_NONE)
    {
        glDeleteTextures(1, &sys->gl_texture);
        sys->gl_texture = GL_NONE;
    }

    /* Cleanup cache */
    if (sys->spu)
        vlc_va_spu_cache_Delete(va, sys->spu);

    /* Cleanup VAAPI resources and connection */
    if (va)
    {
        va->lock();
        vlc_va_conn_DestroySurfaces(va);
        va->unlock();
        vlc_va_Terminate(va);
    }

    /* Cleanup XCB windows */
    if (sys->conn)
    {
        /* show the default cursor */
        xcb_change_window_attributes(sys->conn, sys->embed->handle.xid,
                                     XCB_CW_CURSOR, &(uint32_t) { XCB_CURSOR_NONE });
        xcb_flush(sys->conn);
        xcb_disconnect(sys->conn);
    }

    if (sys->embed)
        vout_display_DeleteWindow (vd, sys->embed);

    if (sys->pool)
    {
        picture_pool_Delete(sys->pool);
        sys->pool = NULL;
    }

    free(vd->sys);

    msg_Info(vd, "VAAPI GLX closed");
}

static void Render(vout_display_t *vd, picture_t *picture, subpicture_t *subpicture)
{
    vout_display_sys_t *sys = vd->sys;

    if (!subpicture)
        return;

    assert(!subpicture->b_ephemer);
    assert(picture->format.i_chroma == VLC_CODEC_VAAPI_SURFACE);

    if (!picture->p_sys)
        return;

    /* count RGBA subpicture regions */
    int i_region = 0;

    subpicture_region_t *region = NULL;
    for (region = subpicture->p_region; region != NULL; region = region->p_next)
    {
        /* Skip non-RGBA regions for now */
        if (region->fmt.i_chroma == VLC_CODEC_RGBA)
            i_region++;
    }

    if (i_region == 0)
        return;

    /* same as previous subpicture? */
    if ((sys->render.prev_subpicture == subpicture) &&
        (sys->render.i_cache > 0))
    {
#ifdef VAAPI_DEBUG
        msg_Dbg(vd, "reuse subpicture %p prev %p cache %d surface %d",
                subpicture, sys->render.prev_subpicture,
                sys->render.i_cache, picture->p_sys->i_id);
#endif
        int ret = vlc_va_spu_cache_Get(VLC_OBJECT(vd), sys->spu, sys->render.i_cache,
                                       picture, subpicture);
        if (ret == VLC_SUCCESS)
            return;
    }

    /* Release exra reference */
    if (sys->render.i_cache > 0)
        vlc_va_spu_cache_Release(sys->va, sys->spu, sys->render.i_cache);

    /* render new subpicture */
    spu_cache_t *cache = vlc_va_spu_cache_Prepare(VLC_OBJECT(vd), sys->va, &sys->sub_fmt,
                                                   picture, subpicture, sys->sflags);
    if (cache == NULL)
        return;

#ifdef VAAPI_DEBUG
    msg_Dbg(vd, "new subpicture %p prev %p cache %d surface %d",
            subpicture, sys->render.prev_subpicture, vlc_va_spu_cache_GetID(cache), surface->i_id);
#endif
    /* remember what was created, so it can be released in DisplayPicture() */
    vlc_va_spu_cache_Put(sys->spu, cache);

    sys->render.i_cache = vlc_va_spu_cache_GetID(cache);
    sys->render.prev_subpicture = subpicture;
    sys->render.i_start = subpicture->i_start;
    sys->render.i_stop  = subpicture->i_stop;

    /* Hold extra reference */
    vlc_va_spu_cache_Hold(sys->spu, vlc_va_spu_cache_GetID(cache));
}

static void DisplayVASurfaceGLX(vout_display_t *vd, picture_t *picture)
{
    vout_display_sys_t *sys = vd->sys;
    vlc_va_conn_t *va = sys->va;

    picture_sys_t *surface = picture->p_sys;
    assert(surface);

    va->lock();

    /* NOTE: libva defines various scaling filters,
     * we use the default here. */
    unsigned int flags = VA_FILTER_SCALING_DEFAULT;

    /* NOTE: vaCreateContext() specifies VA_PROGRESSIVE */
    flags |= picture->b_progressive ? VA_FRAME_PICTURE :
                 (picture->b_top_field_first ? VA_TOP_FIELD : VA_BOTTOM_FIELD);

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, sys->gl_texture);

    VAStatus status;
    status = vaCopySurfaceGLX(va->p_display,
                              sys->gl_surface,
                              surface->i_id,
                              flags);

    glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);

    if (status != VA_STATUS_SUCCESS)
        msg_Err(vd, "failed displaying picture: %d (surface id=%d): %p",
                status, surface->i_id, picture);

    sys->va->unlock();

    /* Render frame */
    glClear(GL_COLOR_BUFFER_BIT);

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, sys->gl_texture);
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    glBegin(GL_QUADS);
    {
        float scale_w, scale_h;
        scale_w = (float)sys->chroma->p[0].w.num / sys->chroma->p[0].w.den / vd->fmt.i_width;
        scale_h = (float)sys->chroma->p[0].h.num / sys->chroma->p[0].h.den / vd->fmt.i_height;

        float left, top, right, bottom;
        left   = (vd->source.i_x_offset + 0) * scale_w;
        top    = (vd->source.i_y_offset + 0) * scale_h;
        right  = (vd->source.i_x_offset + vd->source.i_visible_width) * scale_w;
        bottom = (vd->source.i_y_offset + vd->source.i_visible_height) * scale_h;

        glTexCoord2f(left, top);     glVertex2f(-1.0f,  1.0f);
        glTexCoord2f(right, top);    glVertex2f( 1.0f,  1.0f);
        glTexCoord2f(right, bottom); glVertex2f( 1.0f, -1.0f);
        glTexCoord2f(left, bottom);  glVertex2f(-1.0f, -1.0f);
    }
    glEnd();
    glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);
}

static void DisplayPicture(vout_display_t *vd, picture_t *picture, subpicture_t *subpicture)
{
    vout_display_sys_t *sys = vd->sys;
    vlc_va_conn_t *va = sys->va;

    int ret = VLC_EGENERIC;
    int canc = vlc_savecancel();

    if (!sys->visible)
        goto out;

    assert(picture->format.i_chroma == VLC_CODEC_VAAPI_SURFACE);

    if ((picture->p_sys == NULL) ||
        (picture->p_sys->i_id == VA_INVALID_SURFACE))
    {
        msg_Err(vd, "discarding picture without picture_sys_t information");
        goto out;
    }

    if (subpicture)
    {
        ret = vlc_va_spu_cache_Link(VLC_OBJECT(vd), va, sys->spu, picture, sys->sflags);
        if (ret == VLC_EGENERIC)
            msg_Err(vd, "Failed linking subpicture to surface");
    }

    if (!vlc_gl_Lock(&sys->gl))
    {
        DisplayVASurfaceGLX(vd, picture);

        vlc_gl_Swap(&sys->gl);
        vlc_gl_Unlock(&sys->gl);
    }
    if (vd->cfg->is_fullscreen)
        glClear(GL_COLOR_BUFFER_BIT);

    if (subpicture)
    {
        if (ret == VLC_SUCCESS)
        {
            if (vlc_va_spu_cache_Unlink(VLC_OBJECT(vd), va, sys->spu, picture) == VLC_EGENERIC)
                msg_Err(vd, "Could not find subpicture in cache");
        }
        if (picture->p_sys->i_cache > 0)
            vlc_va_spu_cache_Release(va, sys->spu, picture->p_sys->i_cache);
        picture->p_sys->i_cache = 0;
    }

 out:
    if (subpicture) subpicture_Delete(subpicture);
    picture_Release(picture);
    vlc_restorecancel(canc);
}

static picture_pool_t *Pool(vout_display_t *vd, unsigned requested)
{
    vout_display_sys_t *sys = vd->sys;
    vlc_va_conn_t *va = sys->va;

    if (sys->pool)
        return sys->pool;

    sys->pool = vout_vaapi_GetPool(vd, va, requested);

    xcb_flush(sys->conn);

    return sys->pool;
}
