/**
 * @file vaapi/xcb.c
 * @brief VA API using XCB video output module for VLC media player
 */
/*****************************************************************************
 * Copyright © 2011 - 2013, M2X BV
 * Author: Jean-Paul Saman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <assert.h>

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_es.h>
#include <vlc_picture.h>
#include <vlc_vout_display.h>
#include <vlc_vout_window.h>
#include <vlc_picture_pool.h>

#include <vlc_arrays.h>

#ifdef HAVE_LIBAVCODEC_AVCODEC_H
#   include <libavcodec/avcodec.h>
#   include <libavcodec/vaapi.h>
#else
#   include <avcodec.h>
#   include <vaapi.h>
#endif

#include "../../codec/avcodec/avcodec.h"

#include <xcb/xcb.h>
#include <X11/Xlib-xcb.h>
#include <X11/Xutil.h>
#include <vlc_xlib.h>
#include "../../video_output/xcb/xcb_events_vlc.h"

#include <va/va.h>
#include <va/va_x11.h>
#include "../../codec/avcodec/va.h"
#include "../../codec/avcodec/vaapi.h"

#include "spu.h"
#include "common.h"
#include "xcb.h"

#define VOUT_CFG_PREFIX "vaapi-vout-"

/* define for extra debugging */
#undef VAAPI_DEBUG

static picture_pool_t *Pool(vout_display_t *, unsigned);
static void Render(vout_display_t *, picture_t *, subpicture_t *);
static void DisplayPicture(vout_display_t *, picture_t *, subpicture_t *);

/* */
struct vout_display_sys_t
{
    vlc_va_conn_t    *va;
    xcb_connection_t *conn; /**< XCB connection */

    vout_window_t *embed;

    picture_pool_t *pool;  /* picture pool */
    VAImageFormat img_fmt;
    VAImageFormat sub_fmt; /* Subpicture format VA_FOURCC_RGBA */
    int           sflags;  /* Subtitle format flags */

    struct {
        unsigned int   i_cache;
        mtime_t        i_start;
        mtime_t        i_stop;
        subpicture_t  *prev_subpicture;
    } render;

    vlc_va_spu_cache_t *spu;

    xcb_cursor_t cursor; /* blank cursor */
    xcb_window_t window; /* drawable X window */

    /* whether to draw */
    bool visible;
};

static void Manage(vout_display_t *vd)
{
    int canc = vlc_savecancel();
    vout_display_sys_t *sys = vd->sys;
    ManageEvent(vd, sys->conn, &sys->visible);
    vlc_restorecancel(canc);
}

static int Control(vout_display_t *vd, int query, va_list ap)
{
    vout_display_sys_t *sys = vd->sys;

    switch (query)
    {
    case VOUT_DISPLAY_CHANGE_FULLSCREEN:
    {
        const vout_display_cfg_t *c = va_arg(ap, const vout_display_cfg_t *);
        return vout_window_SetFullScreen(sys->embed, c->is_fullscreen);
    }

    case VOUT_DISPLAY_CHANGE_WINDOW_STATE:
    {
        unsigned state = va_arg(ap, unsigned);
        return vout_window_SetState(sys->embed, state);
    }

    case VOUT_DISPLAY_CHANGE_DISPLAY_SIZE:
    case VOUT_DISPLAY_CHANGE_ZOOM:
    case VOUT_DISPLAY_CHANGE_DISPLAY_FILLED:
    case VOUT_DISPLAY_CHANGE_SOURCE_ASPECT:
    case VOUT_DISPLAY_CHANGE_SOURCE_CROP:
    {
        const vout_display_cfg_t *cfg;
        const video_format_t *source;
        bool is_forced = false;

        if (query == VOUT_DISPLAY_CHANGE_SOURCE_ASPECT
         || query == VOUT_DISPLAY_CHANGE_SOURCE_CROP)
        {
            source = (const video_format_t *)va_arg(ap, const video_format_t *);
            cfg = vd->cfg;
        }
        else
        {
            source = &vd->source;
            cfg = (const vout_display_cfg_t*)va_arg(ap, const vout_display_cfg_t *);
            if (query == VOUT_DISPLAY_CHANGE_DISPLAY_SIZE)
                is_forced = (bool)va_arg(ap, int);
        }

        /* */
        if (query == VOUT_DISPLAY_CHANGE_DISPLAY_SIZE
         && is_forced
         && (cfg->display.width  != vd->cfg->display.width
           ||cfg->display.height != vd->cfg->display.height)
         && vout_window_SetSize(sys->embed,
                                cfg->display.width, cfg->display.height))
            return VLC_EGENERIC;

        vout_display_place_t place;
        vout_display_PlacePicture(&place, source, cfg, false);

        /* Move the picture within the window */
        const uint32_t values[] = { place.x, place.y,
                                    place.width, place.height, };
        xcb_void_cookie_t ck =
            xcb_configure_window_checked(sys->conn, sys->window,
                            XCB_CONFIG_WINDOW_X | XCB_CONFIG_WINDOW_Y
                          | XCB_CONFIG_WINDOW_WIDTH | XCB_CONFIG_WINDOW_HEIGHT,
                              values);
        if (CheckError(vd, sys->conn, "cannot resize XCB window", ck))
            return VLC_EGENERIC;

        return VLC_SUCCESS;
    }
    case VOUT_DISPLAY_RESET_PICTURES:
    {
        assert(0);
        return VLC_SUCCESS;
    }

    /* Hide the mouse. It will be send when
     * vout_display_t::info.b_hide_mouse is false */
    case VOUT_DISPLAY_HIDE_MOUSE:
    {
        xcb_change_window_attributes(sys->conn, sys->embed->handle.xid,
                                     XCB_CW_CURSOR, &(uint32_t){ sys->cursor });
        xcb_flush(sys->conn);
        return VLC_SUCCESS;
    }

    default:
        msg_Err (vd, "Unknown request in VAAPI vout display");
        return VLC_EGENERIC;
    }
}

/**
 * Probe the X server.
 */
int OpenVaapiXCB(vlc_object_t *obj)
{
    vout_display_t *vd = (vout_display_t *) obj;

    const bool b_use_hw = var_GetBool(vd->p_libvlc, "ffmpeg-hw");
    if (!b_use_hw)
    {
        msg_Warn(vd, "requires GPU hardware decoding to be enabled (option --ffmpeg-hw)");
        return VLC_EGENERIC;
    }

    if (!vlc_xlib_init(VLC_OBJECT(vd)))
        return VLC_EGENERIC;

    vout_display_sys_t *sys = calloc(1, sizeof(*sys));
    if (unlikely(sys == NULL))
        return VLC_ENOMEM;

    vd->sys = sys;

    sys->spu = vlc_va_spu_cache_New();
    if (!sys->spu)
    {
        free(sys);
        return VLC_ENOMEM;
    }

    sys->render.prev_subpicture = NULL;
    sys->render.i_cache = 0;

    sys->embed = MakeWindow(vd);
    if (unlikely(sys->embed == NULL))
    {
        CloseVaapiXCB(obj);
        return VLC_EGENERIC;
    }

    /* Create a VA display */
    vlc_va_conn_t *va = vlc_va_Initialize(sys->embed->display.x11);
    if (unlikely(va == NULL))
    {
        CloseVaapiXCB(obj);
        return VLC_EGENERIC;
    }

    /* Connect to X server */
    xcb_connection_t *conn = xcb_connect(sys->embed->display.x11, NULL);
    if (unlikely(xcb_connection_has_error (conn)))
    {
        CloseVaapiXCB(obj);
        return VLC_EGENERIC;
    }

    sys->va = va;
    sys->conn = conn;

    XSetEventQueueOwner(va->x11, XCBOwnsEventQueue);

    RegisterMouseEvents(VLC_OBJECT(vd), conn, sys->embed->handle.xid);

    /* Find window parameters */
    unsigned snum;
    uint8_t depth;
    uint16_t width, height;
    const xcb_screen_t *scr = FindWindow(vd, sys->embed,
                                         conn, &snum, &depth,
                                         &width, &height);
    if (scr == NULL)
    {
        CloseVaapiXCB(obj);
        return VLC_EGENERIC;
    }

    sys->window = xcb_generate_id(conn);

    if (CreateWindow(vd, sys->embed, conn, depth,
                     0 /* ??? */, width, height,
                     &sys->window) != VLC_SUCCESS)
    {
        msg_Err (vd, "cannot create XCB window");
        CloseVaapiXCB(obj);
        return VLC_EGENERIC;
    }

    va->lock();
    video_format_t fmt;
    int result = vlc_va_get_image_format(va, VA_FOURCC( 'Y', 'V', '1', '2' ),
                                         &sys->img_fmt, &fmt);
    va->unlock();

    if (result != VLC_SUCCESS)
    {
        msg_Err(vd, "GPU does not support requested fourcc");
        CloseVaapiXCB(obj);
        return VLC_EGENERIC;
    }

    sys->visible = false;
    sys->cursor = CreateBlankCursor(conn, scr);

    msg_Info(vd, "Video output format size %d x %d / %d x %d",
             vd->fmt.i_width, vd->fmt.i_height,
             vd->fmt.i_visible_width, vd->fmt.i_visible_height);

    vd->fmt.i_chroma = fmt.i_chroma;
    vd->fmt.i_bits_per_pixel = fmt.i_bits_per_pixel;

    /* */
    vout_display_info_t info = vd->info;
    info.is_slow = false;
    info.has_pictures_invalid = false;
    info.has_event_thread = true;
    info.has_hide_mouse = false;
    /* NOTE: Needs to be false for AMD/ATI Graphics cards */
    info.can_recycle = var_CreateGetBool(vd, VOUT_CFG_PREFIX "recycle");

    /* Setup vout_display_t once everything is fine */
    if (VASubtitleFourCC(va, VA_FOURCC_RGBA, &sys->sub_fmt, &sys->sflags) == VLC_SUCCESS)
        info.subpicture_chromas = va_subpicture_chromas;
    else
        msg_Warn(vd, "libva does not support fourcc RGBA subtitles");

    vd->info = info;
    vd->pool = Pool;
    vd->manage = Manage;
    vd->control = Control;
    vd->display = DisplayPicture;
    vd->prepare = vd->info.subpicture_chromas ? Render : NULL;

    /* */
    if (!(sys->sflags & VA_SUBPICTURE_GLOBAL_ALPHA) &&
        !(sys->sflags & VA_SUBPICTURE_CHROMA_KEYING))
        msg_Warn(vd, "Display server does not support alpha or chroma for subtitles.");

    /* */
    bool is_fullscreen = vd->cfg->is_fullscreen;
    if (is_fullscreen && vout_window_SetFullScreen(sys->embed, true))
        is_fullscreen = false;
    vout_display_SendEventFullscreen(vd, is_fullscreen);
    vout_display_SendEventDisplaySize(vd, width, height, is_fullscreen);

    msg_Info(obj, "using VAAPI XCB video output (libva version %d.%d)",
             vd->sys->va->i_version_major, vd->sys->va->i_version_minor);

    return VLC_SUCCESS;
}

void CloseVaapiXCB(vlc_object_t *obj)
{
    vout_display_t *vd = (vout_display_t *) obj;
    vout_display_sys_t *sys = (vout_display_sys_t *) vd->sys;
    vlc_va_conn_t *va = (vlc_va_conn_t *)sys->va;

    sys->visible = false;

    /* Cleanup cache */
    if (sys->spu)
        vlc_va_spu_cache_Delete(va, sys->spu);

    /* */
    if (va)
    {
        va->lock();
        vlc_va_conn_DestroySurfaces(va);
        va->unlock();
        vlc_va_Terminate(va);
    }

    if (sys->conn)
    {
        /* show the default cursor */
        xcb_change_window_attributes(sys->conn, sys->embed->handle.xid,
                                     XCB_CW_CURSOR, &(uint32_t) { XCB_CURSOR_NONE });
        xcb_flush(sys->conn);
        xcb_disconnect(sys->conn);
    }

    if (sys->embed)
        vout_display_DeleteWindow (vd, sys->embed);

    if (sys->pool)
    {
        picture_pool_Delete(sys->pool);
        sys->pool = NULL;
    }

    free(vd->sys);
    msg_Info(obj, "VAAPI XCB closed");
}

static void Render(vout_display_t *vd, picture_t *picture, subpicture_t *subpicture)
{
    vout_display_sys_t *sys = vd->sys;

    if (!subpicture)
        return;

    assert(!subpicture->b_ephemer);
    assert(picture->format.i_chroma == VLC_CODEC_VAAPI_SURFACE);

    if (!picture->p_sys)
        return;

    /* count RGBA subpicture regions */
    int i_region = 0;

    subpicture_region_t *region = NULL;
    for (region = subpicture->p_region; region != NULL; region = region->p_next)
    {
        /* Skip non-RGBA regions for now */
        if (region->fmt.i_chroma == VLC_CODEC_RGBA)
            i_region++;
    }

    if (i_region == 0)
        return;

    /* same as previous subpicture? */
    if ((sys->render.prev_subpicture == subpicture) &&
        (sys->render.i_cache > 0))
    {
#ifdef VAAPI_DEBUG
        msg_Dbg(vd, "reuse subpicture %p prev %p cache %d surface %d",
                subpicture, sys->render.prev_subpicture,
                sys->render.i_cache, picture->p_sys->i_id);
#endif
        int ret = vlc_va_spu_cache_Get(VLC_OBJECT(vd), sys->spu, sys->render.i_cache,
                                       picture, subpicture);
        if (ret == VLC_SUCCESS)
            return;
    }

    /* Release exra reference */
    if (sys->render.i_cache > 0)
        vlc_va_spu_cache_Release(sys->va, sys->spu, sys->render.i_cache);

    /* render new subpicture */
    spu_cache_t *cache = vlc_va_spu_cache_Prepare(VLC_OBJECT(vd), sys->va, &sys->sub_fmt,
                                                  picture, subpicture, sys->sflags);
    if (cache == NULL)
        return;

#ifdef VAAPI_DEBUG
    msg_Dbg(vd, "new subpicture %p prev %p cache %d surface %d",
            subpicture, sys->render.prev_subpicture, vlc_va_spu_cache_GetID(cache), surface->i_id);
#endif
    /* remember what was created, so it can be released in DisplayPicture() */
    vlc_va_spu_cache_Put(sys->spu, cache);

    sys->render.i_cache = vlc_va_spu_cache_GetID(cache);
    sys->render.prev_subpicture = subpicture;
    sys->render.i_start = subpicture->i_start;
    sys->render.i_stop  = subpicture->i_stop;

    /* Hold extra reference */
    vlc_va_spu_cache_Hold(sys->spu, vlc_va_spu_cache_GetID(cache));
}

static void DisplayVASurface(vout_display_t *vd, picture_t *picture)
{
    vout_display_sys_t *sys = vd->sys;
    vlc_va_conn_t *va = sys->va;
    picture_sys_t *surface = picture->p_sys;
    assert(surface);

    sys->va->lock();

    int canc = vlc_savecancel();

    vout_display_place_t place;
    vout_display_PlacePicture(&place, &vd->source, vd->cfg, false);
/*
#define VA_FILTER_SCALING_DEFAULT       0x00000000
#define VA_FILTER_SCALING_FAST          0x00000100
#define VA_FILTER_SCALING_HQ            0x00000200
#define VA_FILTER_SCALING_NL_ANAMORPHIC 0x00000300
#define VA_FILTER_SCALING_MASK          0x00000f00
*/
    unsigned int flags = VA_FILTER_SCALING_DEFAULT;

    /* NOTE: vaCreateContext() specifies VA_PROGRESSIVE */
    flags |= picture->b_progressive ? VA_FRAME_PICTURE :
                 (picture->b_top_field_first ? VA_TOP_FIELD : VA_BOTTOM_FIELD);

    VAStatus status;
    status = vaPutSurface(va->p_display,
                          surface->i_id,
                          sys->window,
                          vd->source.i_x_offset, vd->source.i_y_offset,
                          vd->source.i_visible_width, vd->source.i_visible_height,
                          place.x, place.y, place.width, place.height,
                          NULL, 0, /* clipping */
                          flags);
    if (status != VA_STATUS_SUCCESS)
        msg_Err(vd, "failed displaying picture: %d (surface id=%d): %p",
                status, surface->i_id, picture);

    vlc_restorecancel(canc);

    sys->va->unlock();
}

static void DisplayPicture(vout_display_t *vd, picture_t *picture, subpicture_t *subpicture)
{
    vout_display_sys_t *sys = vd->sys;
    int ret = VLC_EGENERIC;
    int canc = vlc_savecancel();

    if (!sys->visible)
        goto out;

    assert(picture->format.i_chroma == VLC_CODEC_VAAPI_SURFACE);

    if ((picture->p_sys == NULL) ||
        (picture->p_sys->i_id == VA_INVALID_SURFACE))
    {
        msg_Err(vd, "discarding picture without picture_sys_t information");
        goto out;
    }

    if (subpicture)
    {
        ret = vlc_va_spu_cache_Link(VLC_OBJECT(vd), sys->va, sys->spu, picture, sys->sflags);
        if (ret == VLC_EGENERIC)
            msg_Err(vd, "Failed linking subpicture to surface");
    }
    DisplayVASurface(vd, picture);

    if (subpicture)
    {
        if (ret == VLC_SUCCESS)
        {
            if (vlc_va_spu_cache_Unlink(VLC_OBJECT(vd), sys->va, sys->spu, picture) == VLC_EGENERIC)
                msg_Err(vd, "Could not find subpicture in cache");
        }
        if (picture->p_sys->i_cache > 0)
            vlc_va_spu_cache_Release(sys->va, sys->spu, picture->p_sys->i_cache);
        picture->p_sys->i_cache = 0;
    }

 out:
    if (subpicture) subpicture_Delete(subpicture);
    picture_Release(picture);
    vlc_restorecancel(canc);
}

static picture_pool_t *Pool(vout_display_t *vd, unsigned requested)
{
    vout_display_sys_t *sys = vd->sys;
    vlc_va_conn_t *va = sys->va;

    if (sys->pool)
        return sys->pool;

    sys->pool = vout_vaapi_GetPool(vd, va, requested);

    xcb_flush(sys->conn);

    return sys->pool;
}
